import threading
from collections import deque
from copy import copy
from time import sleep, time

from lox import RWLock

SLEEP_TIME = 0.01
N_WORKERS = 5

rw_lock = None
resource = None
resp = None


def common_setup():
    global rw_lock, resp
    rw_lock = RWLock()
    resp = deque()


def common_create_workers(func, n, *args):
    threads = []
    for _ in range(n):
        t = threading.Thread(target=func, args=args)
        threads.append(t)

    t_start = time()
    for t in threads:
        t.start()

    return threads, t_start


def read_worker():
    global rw_lock, resp
    with rw_lock("r"):
        local_copy = copy(resource)
        sleep(SLEEP_TIME)

    resp.append(local_copy)
    return


def write_worker(val):
    global rw_lock, resource
    with rw_lock("w"):
        resource = val
    return


def test_RWLock_r():
    global rw_lock, resource, resp
    common_setup()
    resource = 0

    threads, t_start = common_create_workers(read_worker, N_WORKERS)
    for t in threads:
        t.join()
    t_end = time()
    t_diff = t_end - t_start

    assert N_WORKERS > 2
    # for this to be true, readers have to access at same time (good)
    assert t_diff < (N_WORKERS - 1) * SLEEP_TIME
    for r in resp:
        assert r == resource


def test_RWLock_w():
    global rw_lock, resource
    common_setup()
    resource = 0
    new_val = 5

    threads_w1, t_start_w1 = common_create_workers(write_worker, 1, new_val)

    for t in threads_w1:
        t.join()

    assert resource == new_val


def test_RWLock_rw():
    global rw_lock, resource, resp
    common_setup()

    resource = 0
    soln = [0] * N_WORKERS + [
        5,
    ] * N_WORKERS

    threads_r1, _t_start_r1 = common_create_workers(read_worker, N_WORKERS)
    threads_w1, _t_start_w1 = common_create_workers(write_worker, N_WORKERS, 5)
    threads_r2, _t_start_r2 = common_create_workers(read_worker, N_WORKERS)

    for t in threads_r1:
        t.join()
    for t in threads_w1:
        t.join()
    for t in threads_r2:
        t.join()

    for r, s in zip(resp, soln):
        assert r == s


def test_RWLock_timeout():
    lock = RWLock()

    assert lock.acquire("r", timeout=0.01) is True
    assert lock.acquire("w", timeout=0.01) is False
    assert lock.acquire("r", timeout=0.01) is True
    lock.release("r")
    lock.release("r")

    assert lock.acquire("w", timeout=0.01) is True
    assert lock.acquire("w", timeout=0.01) is False
    assert lock.acquire("r", timeout=0.01) is False
    lock.release("w")


def test_bathroom_example():
    # Note: after the janitor exits, the remaining people are nondeterministic
    sol = [
        "p_0_enter",
        "p_0_exit",
        "j_enter",
        "j_exit",
    ]

    res = bathroom_example()[:4]

    for r, s in zip(res, sol):
        assert r == s


def bathroom_example():
    """
    Scenario:
        A janitor needs to clean a restroom, but is not allowed to enter until
        all people are out of the restroom. How do we implement this?
    """
    restroom = RWLock()
    res = []
    n_people = 5
    sleep_time = 0.1

    def janitor():
        with restroom("w"):  # block until the restroom is no longer occupied
            res.append("j_enter")
            print("(%0.3f s) Janitor  entered the restroom" % (time() - t_start,))
            sleep(sleep_time)  # clean the restroom
            res.append("j_exit")
            print("(%0.3f s) Janitor  exited  the restroom" % (time() - t_start,))

    def people(id):
        if id == 0:  # Get the starting time of execution for display purposes
            global t_start
            t_start = time()
        with restroom("r"):  # block if a janitor is in the restroom
            res.append("p_%d_enter" % (id,))
            print(
                "(%0.3f s) Person %d entered the restroom"
                % (
                    time() - t_start,
                    id,
                )
            )
            sleep(sleep_time)  # use the restroom
            res.append("p_%d_exit" % (id,))
            print(
                "(%0.3f s) Person %d exited  the restroom"
                % (
                    time() - t_start,
                    id,
                )
            )

    people_threads = [
        threading.Thread(target=people, args=(i,)) for i in range(n_people)
    ]
    janitor_thread = threading.Thread(target=janitor)

    for i, person in enumerate(people_threads):
        person.start()  # Person i will now attempt to enter the restroom
        sleep(sleep_time * 0.6)  # wait for 60% the time a person spends in the restroom
        if i == 0:  # While the first person is in the restroom...
            janitor_thread.start()  # the janitor would like to enter. HOWEVER...
            # A new person (until all n_people are done) enters every 0.5 seconds.
    # Wait for all threads to finish
    for t in people_threads:
        t.join()
    janitor_thread.join()

    # The results will look like:
    """
    Running Restroom Demo
    (0.000 s) Person 0 entered the restroom
    (0.100 s) Person 0 exited  the restroom
    (0.101 s) Janitor  entered the restroom
    (0.201 s) Janitor  exited  the restroom
    (0.201 s) Person 1 entered the restroom
    (0.202 s) Person 2 entered the restroom
    (0.202 s) Person 3 entered the restroom
    (0.243 s) Person 4 entered the restroom
    (0.302 s) Person 1 exited  the restroom
    (0.302 s) Person 2 exited  the restroom
    (0.303 s) Person 3 exited  the restroom
    (0.343 s) Person 4 exited  the restroom
    """
    # While Person 0 is in the restroom, Janitor is waiting to enter (at ~0.500 s).
    # While the Janitor is waiting, he doesn't let anyone else into the room.
    # After Person 0, leaves the room, the Janitor enters.
    # After cleaning, the Janitor leaves at the 2.000 second mark.
    # Ever since the janitor was waiting (at 0.500 s), Person 1, Person 2,
    # Person 3, and Person 4 have been lining up to enter.
    # Since the Janitor left the restroom, all waiting people go in at the same time.
    return res


if __name__ == "__main__":
    print("Running Restroom Demo")
    bathroom_example()



import lox
from lox.helper import term_colors


def test_ctx(capsys):
    with capsys.disabled():
        print("")
        with term_colors("red"):
            print("test red")
        with term_colors("green"):
            print("test green")
        with term_colors("blue"):
            print("test blue")
        with term_colors("orange"):
            print("test orange")
        with term_colors("underline"):
            print("test underline")
        with term_colors("bold"):
            print("test bold")
        with term_colors("header"):
            print("test header")



# -*- coding: utf-8 -*-

"""
Objects to help shuttle data beteween tasks.
"""
__all__ = [
    "Announcement",
    "SubscribeFinalizedError",
    "Funnel",
    "FunnelPutError",
    "FunnelPutTopError",
]
from .announcement import Announcement, SubscribeFinalizedError
from .funnel import Funnel, FunnelPutError, FunnelPutTopError



import pytest


def pytest_addoption(parser):
    try:
        parser.addoption(
            "--visual",
            action="store_true",
            default=False,
            help="run interactive visual tests",
        )
    except ValueError:
        pass


def pytest_collection_modifyitems(config, items):
    if config.getoption("--visual"):
        # --visual given in cli: do not skip visual tests
        return
    skip_visual = pytest.mark.skip(reason="need --visual option to run")
    for item in items:
        if "visual" in item.keywords:
            item.add_marker(skip_visual)



import logging as log
import threading
from random import random
from time import sleep, time

import pytest

import lox

SLEEP_TIME = 0.01
N_WORKERS = 5


@pytest.mark.parametrize(
    "n_workers",
    [
        0,
        N_WORKERS,
    ],
)
def test_basic_args(n_workers):
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.thread(n_workers)  # specifying maximum number of threads
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert worker_task(2, 5) == 10

    assert len(worker_task) == 0

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather()

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


def test_basic_noargs():
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.thread  # defaults to a max of 50 threads if none specified
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert 10 == worker_task(2, 5)

    assert len(worker_task) == 0

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather()

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


class Class1:
    def __init__(self, z):
        self.z = z

    @lox.thread(2)
    def test_method1(self, x, y):
        sleep(SLEEP_TIME)
        return x * y + self.z

    @lox.thread
    def test_method2(self, x, y):
        sleep(SLEEP_TIME)
        return x * y - self.z


def test_method_1():
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]
    z1 = 5
    z2 = 5

    test_obj_1 = Class1(z1)
    test_obj_2 = Class1(z2)

    assert (2 * 5 + z1) == test_obj_1.test_method1(2, 5)
    assert (2 * 5 - z1) == test_obj_1.test_method2(2, 5)

    assert (2 * 5 + z2) == test_obj_2.test_method1(2, 5)
    assert (2 * 5 - z2) == test_obj_2.test_method2(2, 5)

    for _ in range(2):
        for x, y in zip(in_x, in_y):
            test_obj_1.test_method1.scatter(x, y)
        res = test_obj_1.test_method1.gather()
        assert len(res) == len(in_x)
        for r, x, y in zip(res, in_x, in_y):
            assert (x * y + z1) == r

    for x, y in zip(in_x, in_y):
        test_obj_2.test_method1.scatter(x, y)
    res = test_obj_2.test_method1.gather()
    assert len(res) == len(in_x)
    for r, x, y in zip(res, in_x, in_y):
        assert (x * y + z2) == r

    for x, y in zip(in_x, in_y):
        test_obj_1.test_method2.scatter(x, y)
    res = test_obj_1.test_method2.gather()
    assert len(res) == len(in_x)
    for r, x, y in zip(res, in_x, in_y):
        assert (x * y - z1) == r

    for x, y in zip(in_x, in_y):
        test_obj_2.test_method2.scatter(x, y)
    res = test_obj_2.test_method2.gather()
    assert len(res) == len(in_x)
    for r, x, y in zip(res, in_x, in_y):
        assert (x * y - z2) == r


@pytest.fixture
def mock_tqdm(mocker):
    return mocker.patch("lox.worker.thread.TQDM")


def test_tqdm_bool(mock_tqdm):

    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.thread(N_WORKERS)  # specifying maximum number of threads
    def worker_task(x, y):
        sleep(random())
        return x * y

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather(tqdm=True)

    mock_tqdm.assert_called_once_with(total=len(in_x))

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


def test_tqdm_tqdm(mocker):
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.thread(N_WORKERS)  # specifying maximum number of threads
    def worker_task(x, y):
        sleep(random())
        return x * y

    pbar = mocker.MagicMock()
    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather(tqdm=pbar)

    assert pbar.update.call_count == len(in_x)

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


@pytest.mark.visual
def test_tqdm_bool_visual():
    """Primarily for visually asserting our tqdm mocks are correct."""

    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.thread(N_WORKERS)  # specifying maximum number of threads
    def worker_task(x, y):
        sleep(random())
        return x * y

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather(tqdm=True)

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


def test_non_decorator():
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    def worker_task_private(x, y):
        sleep(SLEEP_TIME)
        return x * y

    worker_task = lox.thread(worker_task_private)

    # Vanilla function execution still works
    assert worker_task(2, 5) == 10

    assert len(worker_task) == 0

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather()

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


def test_non_decorator_specify_num_workers():
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    def worker_task_private(x, y):
        sleep(SLEEP_TIME)
        return x * y

    worker_task = lox.thread(N_WORKERS)(worker_task_private)

    # Vanilla function execution still works
    assert worker_task(2, 5) == 10

    assert len(worker_task) == 0

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather()

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r



# -*- coding: utf-8 -*-

"""
Concurrency control objects to help parallelized tasks communicate and share resources.
"""

__all__ = [
    "IndexSemaphore",
    "LightSwitch",
    "QLock",
    "RWLock",
]

from .index_semaphore import IndexSemaphore
from .light_switch import LightSwitch
from .qlock import QLock
from .rw_lock import RWLock



import queue
from threading import Lock, Thread
from time import sleep

import pytest

import lox
from lox import Announcement


def test_1():
    """Test most basic usecase."""

    ann = lox.Announcement()
    x_in = [1, 2, 3, 4, 5]
    foo_soln, bar_soln = [], []

    foo_q = ann.subscribe()
    bar_q = ann.subscribe()

    assert isinstance(foo_q, lox.Announcement)
    assert isinstance(bar_q, lox.Announcement)
    assert foo_q != bar_q

    def foo():
        x = foo_q.get()
        foo_soln.append(x**2)

    def bar():
        x = bar_q.get()
        bar_soln.append(x**3)

    threads = []
    for _ in x_in:
        threads.append(Thread(target=foo))
        threads.append(Thread(target=bar))

    for x in x_in:
        ann.put(x)

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    assert len(foo_soln) == len(x_in)
    assert len(bar_soln) == len(x_in)

    for x, r in zip(x_in, foo_soln):
        assert r == x**2

    for x, r in zip(x_in, bar_soln):
        assert r == x**3


def test_2():
    """Test the backlog."""

    ann = lox.Announcement(backlog=0)
    x_in = [1, 2, 3, 4, 5]
    foo_soln, bar_soln = [], []

    foo_q = ann.subscribe()

    def foo():
        x = foo_q.get()
        foo_soln.append(x**2)

    def bar():
        x = bar_q.get()
        bar_soln.append(x**3)

    threads = []
    for _ in x_in:
        threads.append(Thread(target=foo))
        threads.append(Thread(target=bar))

    for x in x_in:
        ann.put(x)

    bar_q = ann.subscribe()
    ann.finalize()

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    assert len(foo_soln) == len(x_in)
    assert len(bar_soln) == len(x_in)

    for x, r in zip(x_in, foo_soln):
        assert r == x**2

    for x, r in zip(x_in, bar_soln):
        assert r == x**3


def test_3():
    """Test subscribing after finalize."""

    ann = lox.Announcement()
    ann.finalize()
    with pytest.raises(lox.SubscribeFinalizedError):
        ann.subscribe()


def test_4():
    """Testing many-to-many capability."""

    ann_foo = Announcement()
    ann_bar = ann_foo.subscribe()
    ann_baz = ann_bar.subscribe()

    ann_foo.put(1)

    assert ann_bar.get() == 1
    assert ann_baz.get() == 1

    ann_bar.put(2)
    assert ann_foo.get() == 2
    assert ann_baz.get() == 2

    ann_bar.put(3)
    assert ann_foo.get() == 3
    assert ann_baz.get() == 3


def test_5():
    """Testing many-to-many backlog and finalization."""

    ann_foo = lox.Announcement(backlog=-1)
    ann_bar = ann_foo.subscribe()

    ann_foo.put(1)
    ann_foo.put(3)
    ann_foo.put(7)

    assert ann_bar.get() == 1
    assert ann_bar.get() == 3
    assert ann_bar.get() == 7

    ann_baz = ann_bar.subscribe()

    assert ann_baz.get() == 1
    assert ann_baz.get() == 3
    assert ann_baz.get() == 7

    ann_boo = ann_bar.subscribe()
    ann_boo.finalize()

    assert ann_boo.get() == 1
    assert ann_boo.get() == 3
    assert ann_boo.get() == 7

    with pytest.raises(queue.Empty):
        ann_boo.get(timeout=0.01)

    with pytest.raises(lox.SubscribeFinalizedError):
        ann_foo.subscribe()



"""
.. module:: process
   :synopsis: Easily execute a function or method in multiple processes.

Still allows the decorated function/method as normal.

Example
-------
    .. doctest::
        :skipif: True

        >>> import lox
        >>>
        >>> @lox.process(4)  # Will operate with a maximum of 4 processes
        ... def foo(x, y):
        ...     return x * y
        ...
        >>> foo(3, 4)
        12
        >>> for i in range(5):
        ...     foo.scatter(i, i + 1)
        ...
        >>> # foo is currently being executed in 4 processes
        >>> results = foo.gather()
        >>> print(results)
        [0, 2, 6, 12, 20]
"""

import os
import threading
from collections import deque
from typing import Callable

try:
    import pathos.multiprocessing as mp

    _PATHOS_AVAILABLE = True
except ImportError:
    mp = None
    _PATHOS_AVAILABLE = False

from ..debug import LOX_DEBUG

try:
    from tqdm import tqdm as TQDM  # to avoid argument namespace collisions.
except ModuleNotFoundError:
    TQDM = None


__all__ = [
    "process",
]


class ScatterGatherCallable:
    def __init__(self, fn, instance, executor, pending, n_workers):
        self._fn = fn
        self._instance = instance
        self._executor = executor
        self._pending = pending
        self._n_workers = n_workers
        self._results_thread_lock = threading.Lock()
        self._tqdm = None
        self._tqdm_pre_update = 0  # Stores updates before we have a tqdm object.

    def __call__(self, *args, **kwargs):
        if self._instance is not None:
            args = (self._instance,) + args
        return self._fn(*args, **kwargs)

    def scatter(self, *args, **kwargs):
        if not _PATHOS_AVAILABLE:
            raise ValueError(
                "pathos is not installed. "
                "Install the full package with: pip install lox[multiprocessing]"
            )

        if self._instance is not None:
            args = (self._instance,) + args

        if self._executor[0] is None:
            # We create the pool here for greater serialization compatability
            self._executor[0] = mp.Pool(self._n_workers)

        def callback(res):
            # All of these are executed in a single "results" thread
            with self._results_thread_lock:
                if self._tqdm is not None:
                    self._tqdm.update(1)
                else:
                    self._tqdm_pre_update += 1

        fut = self._executor[0].apply_async(
            self._fn, args=args, kwds=kwargs, callback=callback
        )
        self._pending.append(fut)

        return fut

    def gather(self, *, tqdm=None):
        if not _PATHOS_AVAILABLE:
            raise ValueError(
                "pathos is not installed. "
                "Install the full package with: pip install lox[multiprocessing]"
            )

        if tqdm is not None:
            if TQDM is None:
                raise ModuleNotFoundError("No module named 'tqdm'")

            if isinstance(tqdm, bool) and tqdm:
                tqdm = TQDM(total=len(self._pending))

            with self._results_thread_lock:
                self._tqdm = tqdm
                # Update the progressbar with all of the results from before
                # the TQDM object was declared.
                self._tqdm.update(self._tqdm_pre_update)

        self._executor[0].close()
        self._executor[0].join()
        fetched = [x.get() for x in self._pending]
        self._pending.clear()
        self._executor[0] = None
        return fetched


class ScatterGatherDescriptor:
    def __init__(self, fn, n_workers):
        self._executor = [None]  # Make a list to share a "pointer"
        self._n_workers = n_workers
        self._fn = fn
        self._pending = deque()
        self._base_callable = ScatterGatherCallable(
            self._fn, None, self._executor, self._pending, self._n_workers
        )

    def __call__(self, *args, **kwargs):
        """
        Vanilla passthrough function execution. Default user function behavior.

        Returns
        -------
        Decorated function return type.
           Return of decorated function.
        """

        return self._fn(*args, **kwargs)

    def __len__(self):
        """Return length of unprocessed job queue.

        Returns
        -------
            Approximate length of unprocessed job queue.
        """

        count = 0
        for res in self._pending:
            if not res.ready():
                count += 1

        return count

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return ScatterGatherCallable(
            self._fn, instance, self._executor, self._pending, self._n_workers
        )

    def scatter(self, *args, **kwargs):
        """Enqueue a job to be processed by workers.
        Spin up workers if necessary.

        Return
        ------
        concurrent.futures.Future
        """
        if not _PATHOS_AVAILABLE:
            raise ValueError(
                "pathos is not installed. "
                "Install the full package with: pip install lox[multiprocessing]"
            )

        return self._base_callable.scatter(*args, **kwargs)

    def gather(self, *args, **kwargs):
        """Block and collect results from prior ``scatter`` calls."""
        if not _PATHOS_AVAILABLE:
            raise ValueError(
                "pathos is not installed. "
                "Install the full package with: pip install lox[multiprocessing]"
            )

        results = self._base_callable.gather(*args, **kwargs)
        self._executor = None
        return results


def process(n_workers) -> Callable[[Callable], ScatterGatherDescriptor]:
    """Decorate a function/method to execute in multiple processes.

    Example
    -------
    .. doctest::
        :skipif: True

        >>> import lox
        >>>
        >>> @lox.process(4)  # Will operate with a maximum of 4 processes
        ... def foo(x, y):
        ...     return x * y
        ...
        >>> foo(3, 4)
        12
        >>> for i in range(5):
        ...     foo.scatter(i, i + 1)
        ...
        >>> # foo is currently being executed in 4 processes
        >>> results = foo.gather()
        >>> print(results)
        [0, 2, 6, 12, 20]

    Parameters
    ----------
    n_workers : int
        Number of process workers to invoke. Defaults to number of CPU cores.

    Methods
    -------
    __call__( *args, **kwargs )
        Vanilla passthrough function execution. Default user function behavior.

        Returns
        -------
        Decorated function return type.
            Return of decorated function.
    __len__()
        Returns
        -------
        int
            job queue length.
    scatter( *args, **kwargs )
        Start a job executing ``func( *args, **kwargs )``.
        Workers are created and destroyed automatically.

        Returns
        -------
        int
            Solution's index into the results obtained via ``gather()``.
    gather()
        Block until all jobs called via ``scatter()`` are complete.

        Returns
        -------
        list
            Results in the order that scatter was invoked.
    """  # noqa: D214,D215,D410,D411

    # Support @process with no arguments.
    if callable(n_workers):
        return process(os.cpu_count())(n_workers)

    def decorator(fn: Callable) -> ScatterGatherDescriptor:
        return ScatterGatherDescriptor(fn, 0 if LOX_DEBUG else n_workers)

    return decorator


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
.. module:: helper
   :synopsis: Private general lox helper functions/classes
"""

__all__ = [
    "term_colors",
]


class term_colors:
    """Escape sequences to colorize text in the terminal"""

    HEADER = "\033[95m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    ORANGE = "\033[93m"
    RED = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    def __init__(self, color: str):
        self.color = color.lower()

    def __enter__(
        self,
    ):
        if self.color == "red":
            print(self.RED, end="")
        elif self.color == "green":
            print(self.GREEN, end="")
        elif self.color == "blue":
            print(self.BLUE, end="")
        elif self.color == "orange" or self.color == "yellow":
            print(self.ORANGE, end="")
        elif self.color == "bold":
            print(self.BOLD, end="")
        elif self.color == "underline":
            print(self.UNDERLINE, end="")
        elif self.color == "header":
            print(self.HEADER, end="")
        else:
            raise ValueError("Invalid color string")

    def __exit__(self, type, value, traceback):
        print(self.ENDC, end="", flush=True)



"""
.. module:: thread
   :synopsis: Easily execute a function or method in multiple threads.

Still allows the decorated function/method as normal.

Example
-------
.. doctest::

    >>> import lox
    >>>
    >>> @lox.thread(4)  # Will operate with a maximum of 4 threads
    ... def foo(x, y):
    ...     return x * y
    ...
    >>> foo(3, 4)
    12
    >>> for i in range(5):
    ...     foo.scatter(i, i + 1)
    ...
    -ignore-
    >>> # foo is currently being executed in 4 threads
    >>> results = foo.gather()  # block until results are ready
    >>> print(results)  # Results are in the same order as scatter() calls
    [0, 2, 6, 12, 20]
"""

import concurrent.futures
import threading
from typing import Callable

from ..debug import LOX_DEBUG

try:
    from tqdm import tqdm as TQDM  # to avoid argument namespace collisions.
except ModuleNotFoundError:
    TQDM = None


__all__ = [
    "thread",
]


class ScatterGatherCallable:
    def __init__(self, fn, instance, executor, pending, pending_lock):
        self._fn = fn
        self._instance = instance
        self._executor = executor
        self._pending = pending
        self._pending_lock = pending_lock

    def __call__(self, *args, **kwargs):
        if self._instance is not None:
            args = (self._instance,) + args
        return self._fn(*args, **kwargs)

    def scatter(self, *args, **kwargs):
        if self._instance is not None:
            args = (self._instance,) + args
        with self._pending_lock:
            if self._executor:
                fut = self._executor.submit(self._fn, *args, **kwargs)
            else:
                # Execute function in current thread
                fut = concurrent.futures.Future()
                fut.set_running_or_notify_cancel()
                try:
                    res = self._fn(*args, **kwargs)
                except Exception as e:
                    fut.set_exception(e)
                else:
                    fut.set_result(res)
            self._pending.append(fut)
        return fut

    def gather(self, *, tqdm=None):
        """
        Parameters
        ----------
        tqdm : tqdm.tqdm or bool
            If ``True``, will internally create and update a default tqdm object.
            If ``tqdm``, initialized tqdm object to update with progress.
        """

        pending = []
        with self._pending_lock:
            pending[:] = self._pending
            self._pending[:] = []

        if tqdm is not None:
            if TQDM is None:
                raise ModuleNotFoundError("No module named 'tqdm'")

            if isinstance(tqdm, bool) and tqdm:
                tqdm = TQDM(total=len(pending))

            # We need a mutex for the tqdm object since multiple callbacks
            # can be called at the same time via different threads under
            # the same parent process.
            tqdm_mutex = threading.Lock()

            def tqdm_callback(fut):
                with tqdm_mutex:
                    tqdm.update(1)

            [fut.add_done_callback(tqdm_callback) for fut in pending]

        output = [fut.result() for fut in pending]

        return output


class ScatterGatherDescriptor:
    def __init__(self, fn, n_workers):
        """
        Note: define self._executor in child class before calling this
        """

        self._pending_lock = threading.Lock()
        self._fn = fn
        self._pending = []

        if n_workers:
            self._executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=n_workers
            )
        else:
            # Disable threading; scatter calls are blocking; useful
            # for debugging multithreaded functions.
            self._executor = None

        self._base_callable = ScatterGatherCallable(
            self._fn, None, self._executor, self._pending, self._pending_lock
        )

    def __call__(self, *args, **kwargs):
        """
        Vanilla passthrough function execution. Default user function behavior.

        Returns
        -------
        Decorated function return type.
           Return of decorated function.
        """

        return self._fn(*args, **kwargs)

    def __len__(self):
        """Return length of unprocessed job queue.

        Returns
        -------
            Approximate length of unprocessed job queue.
        """

        if self._executor:
            return self._executor._work_queue.qsize()
        else:
            return 0

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return ScatterGatherCallable(
            self._fn, instance, self._executor, self._pending, self._pending_lock
        )

    def scatter(self, *args, **kwargs):
        """Enqueue a job to be processed by workers.
        Spin up workers if necessary.

        Return
        ------
        concurrent.futures.Future
        """

        return self._base_callable.scatter(*args, **kwargs)

    def gather(self, *args, **kwargs):
        """Block and collect results from prior ``scatter`` calls."""

        return self._base_callable.gather(*args, **kwargs)


def thread(n_workers) -> Callable[[Callable], ScatterGatherDescriptor]:
    """Decorate a function/method to execute in multiple threads.

    Example:

    .. doctest::

        >>> import lox
        >>>
        >>> @lox.thread(4)  # Will operate with a maximum of 4 threads
        ... def foo(x, y):
        ...     return x * y
        ...
        >>> foo(3, 4)
        12
        >>> for i in range(5):
        ...     foo.scatter(i, i + 1)
        ...
        -ignore-
        >>> # foo is currently being executed in 4 threads
        >>> results = foo.gather()
        >>> print(results)
        [0, 2, 6, 12, 20]

    Multiple decorated functions can be chained together, each function drawing
    from their own pool of threads. Functions that return tuples will automatically
    unpack into the chained function. Positional arguments and keyword arguments
    can be passed in as they normally would.

    .. doctest::
        >>> @lox.thread(2)  # Will operate with a maximum of 2 threads
        ... def bar(x, y):
        ...     return x + y
        ...

        >>> for i in range(5):
        ...     foo_res = foo.scatter(i, i + 1)
        ...     bar.scatter(foo_res, 10)  # scatter will automatically
        ...     # unpack the results of foo
        ...
        >>>
        >>> results = bar.gather()

    Parameters
    ----------
    max_workers : int
        Maximum number of threads to invoke.
        When ``lox.thread`` is called without ``()``, the wrapped function
        a default number of max_workers is used (50).
        If set to 0, scatter calls will be executed in the parent thread.

    Methods
    -------
    __call__( *args, **kwargs )
        Vanilla passthrough function execution. Default user function behavior.

        Returns
        -------
        Decorated function return type.
           Return of decorated function.

    __len__()
        Returns
        -------
        int
            Current job queue length. Number of jobs that are currently waiting
            for an available worker.

    scatter( *args, **kwargs)
        Start a job executing decorated function ``func( *args, **kwargs )``.
        Workers are created and destroyed automatically.

    gather()
        Block until all jobs called via ``scatter()`` are complete.

        Returns
        -------
        list
            Results in the order that scatter was invoked.
    """  # noqa: D214,D215,D410,D411

    # Support @thread with no arguments.
    if callable(n_workers):
        return thread(50)(n_workers)

    def decorator(fn: Callable) -> ScatterGatherDescriptor:
        return ScatterGatherDescriptor(fn, 0 if LOX_DEBUG else n_workers)

    return decorator


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from random import random
from time import sleep, time

import pytest

import lox

SLEEP_TIME = 0.01
N_WORKERS = 2


@pytest.mark.requires_pathos
def test_basic_args():
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.process(N_WORKERS)  # specifying maximum number of processes
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert 10 == worker_task(2, 5)

    assert len(worker_task) == 0
    print("test_process: %s" % (worker_task))

    for _ in range(2):
        for x, y in zip(in_x, in_y):
            worker_task.scatter(x, y)

        res = worker_task.gather()
        assert len(res) == len(in_x)

        for r, x, y in zip(res, in_x, in_y):
            assert (x * y) == r


@pytest.mark.requires_pathos
def test_basic_noargs():
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.process  # maximum number of processes = # cores
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert 10 == worker_task(2, 5)

    assert len(worker_task) == 0
    print("test_process: %s" % (worker_task))

    for _ in range(2):
        for x, y in zip(in_x, in_y):
            worker_task.scatter(x, y)

        res = worker_task.gather()
        assert len(res) == len(in_x)

        for r, x, y in zip(res, in_x, in_y):
            assert (x * y) == r


class Class1:
    def __init__(self, z):
        self.z = z

    @lox.process(2)
    def test_method1(self, x, y):
        sleep(SLEEP_TIME)
        return x * y + self.z

    @lox.process
    def test_method2(self, x, y):
        sleep(SLEEP_TIME)
        return x * y + self.z


@pytest.mark.requires_pathos
def test_method_1():
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]
    z = 5

    test_obj = Class1(z)

    assert (2 * 5 + z) == test_obj.test_method1(2, 5)
    assert (2 * 5 + z) == test_obj.test_method2(2, 5)

    for i in range(2):
        for x, y in zip(in_x, in_y):
            test_obj.test_method1.scatter(x, y)
        res = test_obj.test_method1.gather()
        assert len(res) == len(in_x), f"iteration {i}"

        for r, x, y in zip(res, in_x, in_y):
            assert (x * y + z) == r

        for x, y in zip(in_x, in_y):
            test_obj.test_method2.scatter(x, y)
        res = test_obj.test_method2.gather()
        assert len(res) == len(in_x), f"iteration {i}"

        for r, x, y in zip(res, in_x, in_y):
            assert (x * y + z) == r


@pytest.fixture
def mock_tqdm(mocker):
    return mocker.patch("lox.worker.process.TQDM")


@pytest.mark.requires_pathos
def test_tqdm_bool(mock_tqdm):

    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.process(N_WORKERS)  # specifying maximum number of processes
    def worker_task(x, y):
        sleep(random())
        return x * y

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather(tqdm=True)

    mock_tqdm.assert_called_once_with(total=len(in_x))

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


@pytest.mark.requires_pathos
def test_tqdm_tqdm(mocker):
    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.process(N_WORKERS)  # specifying maximum number of processes
    def worker_task(x, y):
        sleep(random())
        return x * y

    pbar = mocker.MagicMock()
    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather(tqdm=pbar)

    n_update = 0
    for args, _kwargs in pbar.update.call_args_list:
        n_update += args[0]
    assert n_update == len(in_x)

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r


@pytest.mark.visual
@pytest.mark.requires_pathos
def test_tqdm_bool_visual():
    """Primarily for visually asserting our tqdm mocks are correct."""

    in_x = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    in_y = [
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
    ]

    @lox.process(N_WORKERS)  # specifying maximum number of processes
    def worker_task(x, y):
        sleep(random())
        return x * y

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather(tqdm=True)

    assert len(res) == len(in_x)

    for r, x, y in zip(res, in_x, in_y):
        assert (x * y) == r



# -*- coding: utf-8 -*-

"""
Easy multithreading for every project.
"""

__author__ = """Brian Pugh"""
__email__ = "bnp117@gmail.com"
__version__ = "1.0.0"

__all__ = [
    "Announcement",
    "Funnel",
    "FunnelPutError",
    "FunnelPutTopError",
    "IndexSemaphore",
    "LOX_DEBUG",
    "LightSwitch",
    "QLock",
    "RWLock",
    "SubscribeFinalizedError",
    "Timeout",
    "process",
    "thread",
]

import lox.worker
from lox.worker.process import process
from lox.worker.thread import thread

from .debug import LOX_DEBUG
from .exceptions import Timeout
from .lock import IndexSemaphore, LightSwitch, QLock, RWLock
from .queue import (
    Announcement,
    Funnel,
    FunnelPutError,
    FunnelPutTopError,
    SubscribeFinalizedError,
)



import queue
from threading import Lock, Thread
from time import sleep

import lox
from lox import Announcement


def test_1():
    """Test most basic usecase with 1 subscribers."""
    funnel = lox.Funnel()
    sub_1 = funnel.subscribe()

    sub_1.put("foo", "job_id")
    res = funnel.get()
    assert len(res) == 2
    assert res[0] == "job_id"
    assert res[1] == "foo"
    assert funnel.d == {}


def test_2():
    """Test most basic usecase with 2 subscribers."""

    funnel = lox.Funnel()
    sub_1 = funnel.subscribe()
    sub_2 = funnel.subscribe()

    sub_1.put("foo", "job_id")
    try:
        res = funnel.get(timeout=0.01)
        assert 0
    except queue.Empty:
        assert 1

    sub_2.put("bar", "job_id")

    res = funnel.get()
    assert len(res) == 3
    assert res[0] == "job_id"
    assert res[1] == "foo"
    assert res[2] == "bar"
    assert funnel.d == {}



import os
import sys

from setuptools import find_packages, setup
from setuptools.command.install import install

version = "1.0.0"


with open("README.rst") as readme_file:
    readme = readme_file.read()
readme = readme.replace(
    "assets/lox_200w.png",
    "https://raw.githubusercontent.com/BrianPugh/lox/main/assets/lox_200w.png",
)

with open("HISTORY.rst") as history_file:
    history = history_file.read()

setup(
    author="Brian Pugh",
    author_email="bnp117@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    description="Threading and Multiprocessing for every project.",
    install_requires=[],
    extras_require={
        "multiprocessing": ["pathos"],
    },
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="lox",
    name="lox",
    packages=find_packages(include=["lox"]),
    test_suite="tests",
    url="https://github.com/BrianPugh/lox",
    version=version,
    zip_safe=False,
)
